# Licensed under the GPL: https://www.gnu.org/licenses/gpl-3.0.en.html
# For details: reprotest/debian/copyright

import sys
from reprotest import main

if __name__ == '__main__':
    sys.exit(main())
